#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

# Docstring essential for docopt arguments
"""
PyPEF - Pythonic Protein Engineering Framework.

Creation of learning and validation sets: to split .CSV data in Learning and Validation sets run
    run_pypef.py mklsvs [...]
Creation of prediction sets: To create prediction sets from CSV data single point mutational variants run
    run_pypef.py mkps [...]
Running:
 1. To train and validate models run
        run_pypef.py run -l Learning_Set.fasta -v Validation_Set.fasta [-s 5] [--parallel] [-c 4]
    ! Attention using ray for parallel computing ('--parallel') in Windows: Ray is not yet fully supported for Windows !
 2. To plot the validation creating a figure (.png) run
        run_pypef.py run -m MODEL12345 -f Validation_Set.fasta
 3. To predict variants run
        run_pypef.py run -m MODEL12345 -p Prediction_Set.fasta
    or for predicting variants in created prediction set folders exemplary run
        run_pypef.py run -m MODEL12345 --pmult [--drecomb] [...] [--qdiverse]
    or for performing in silico directed evolution run:
        run_pypef.py directevo -m MODEL12345 [...]


Usage:
    run_pypef.py mklsvs [--wtseq WT_SEQ] [--input CSV_FILE] [--drop THRESHOLD] [--nornd NUMBER]
    run_pypef.py mkps [--wtseq WT_SEQ] [--input CSV_FILE] [--drop THRESHOLD]
                                 [--drecomb] [--trecomb] [--qrecomb]
                                 [--ddiverse] [--tdiverse] [--qdiverse]
    run_pypef.py run --ls LEARNING_SET --vs VALIDATION_SET [--save NUMBER] [--regressor TYPE] [--nofft] [--all]
                                                       [--sort METRIC] [--parallel] [--cores NUMCORES]
    run_pypef.py --show [MODELS]
    run_pypef.py run --model MODEL12345 --figure VS_FOR_PLOTTING  [--label] [--color] [--ywt WT_FITNESS] [--nofft]
    run_pypef.py run --model MODEL12345 --ps PREDICTION_SET [--nofft] [--negative] [--print]
    run_pypef.py run --model MODEL12345 --pmult [--drecomb] [--trecomb] [--qrecomb]
                                          [--ddiverse] [--tdiverse] [--qdiverse] [--nofft] [--negative]
    run_pypef.py directevo --model MODEL12345 [--ywt WT_FITNESS] [--wtseq WT_SEQ]
                                        [--numiter NUM_ITER] [--numtraj NUM_TRAJ]
                                        [--temp TEMPERATURE] [--nofft] [--negative] [--print]
                                        [--usecsv] [--csvaa] [--input CSV_FILE] [--drop THRESHOLD]


Options:
  -h --help                    Show this screen.
  --version                    Show version.
  --show                       Show saved models in Model_Results.txt.
  MODELS                       Number of saved models to show [default: 5].
  -w --wtseq WT_SEQ            Input file (in .fa format) for wild-type sequence [default: None].
  -i --input CSV_FILE          Input data file in .csv format [default: None].
  -d --drop THRESHOLD          Below threshold variants will be discarded from the data [default: -9E09].
  -n --nornd NUMBER            Number of randomly created Learning and Validation datasets [default: 0].
  -s --save NUMBER             Number of Models to be saved as pickle files [default: 5].
  --parallel                   Parallel computing of training and validation of models [default: False].
  -c --cores NUMCORES          Number of cores used in parallel computing.
  --drecomb                    Create/predict double recombinants [default: False].
  --trecomb                    Create/predict triple recombinants [default: False].
  --qrecomb                    Create/predict quadruple recombinants [default: False].
  --ddiverse                   Create/predict double natural diverse variants [default: False].
  --tdiverse                   Create/predict triple natural diverse variants [default: False].
  --qdiverse                   Create quadruple natural diverse variants [default: False].
  -u --pmult                   Predict for all prediction files in folder for recomb
                               or diverse variants [default: False].
  --negative                   Set if more negative values define better variants [default: False].
  -l --ls LEARNING_SET         Input learning set in .fasta format.
  -v --vs VALIDATION_SET       Input validation set in .fasta format.
  --regressor TYPE             Type of regression (R.) to use, options: PLS R.: pls, PLS CV R.: pls_cv,
                               Random Forest R.: rf, SVM R.: svr, MLP R.: mlp [default: pls].
  --nofft                      Raw sequence input, i.e. no FFT for establishing protein spectra
                               as vector inputs [default: False].
  --all                        Finally training on all data [default: False]
  --sort METRIC                Rank models based on metric {1: R^2, 2: RMSE, 3: NRMSE, 4: Pearson's r,
                               5: Spearman's rank} [default: 1].
  -m --model MODEL12345        Model (pickle file) for plotting of validation or for performing predictions.
  -f --figure VS_FOR_PLOTTING  Validation set for plotting using a trained model.
  --label                      Label the plot instances [default: False].
  --color                      Color the plot for "true" and "false" predictions quarters [default: False].
  -p --ps PREDICTION_SET       Prediction set for performing predictions using a trained Model.
  --print                      Print raw encoded and FFT-ed sequence matrices and predictions in Shell [default: False].
  -y --ywt WT_FITNESS          Fitness value (y) of wild-type.
  --numiter NUM_ITER           Number of mutation iterations per evolution trajectory [default: 5].
  --numtraj NUM_TRAJ           Number of trajectories, i.e. evolution pathways [default: 5].
  --temp TEMPERATURE           "Temperature" of Metropolis-Hastings criterion [default: 0.01]
  --usecsv                     Perform directed evolution on single variant csv position data [default: False].
  --csvaa                      Directed evolution csv amino acid substitutions,
                               requires flag "--usecsv" [default: False].
"""

# standard import, for all required modules see requirements.txt file(s)
import os
from os import listdir
from os.path import isfile, join
import numpy as np
from tqdm import tqdm
from docopt import docopt
import multiprocessing
# ray imported later locally as is is only used for parallelized running, thus commented out:
# import ray

# importing own modules
from pypef.Modules_Regression import (
    read_models, formatted_output, r2_list, save_model, predict,
    predictions_out, plot
)
from pypef.Modules_Directed_Evolution import run_de_trajectories
from pypef.Modules_Learning_Validation import (
    get_wt_sequence, csv_input, drop_rows, get_variants, make_sub_ls_vs,
    make_sub_ls_vs_randomly, make_fasta_ls_vs
)
from pypef.Modules_Prediction import (
    make_combinations_double, make_combinations_triple, make_combinations_quadruple,
    create_split_files, make_combinations_double_all_diverse,
    make_combinations_triple_all_diverse, make_combinations_quadruple_all_diverse
)
# import Modules_Parallelization.r2_list_parallel locally to avoid error
# when not running in parallel, thus commented out:
# from pypef.Modules_Parallelization import r2_list_parallel


def run():
    """
    Running the program, importing all required self-made modules and
    running them dependent on user-passed input arguments using docopt
    for argument parsing.
    """
    arguments = docopt(__doc__, version='PyPEF 0.1.3 (August 2021)')
    # print(arguments)  # uncomment for printing parsed docopt arguments
    amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']

    if arguments['--show']:
        if arguments['MODELS'] != str(5):
            try:
                print(read_models(int(arguments['MODELS'])))
            except ValueError:
                print(read_models(5))
            except TypeError:
                print(read_models(5))
        else:
            print(read_models(5))

    if arguments['mklsvs']:
        wt_sequence = get_wt_sequence(arguments['--wtseq'])
        csv_file = csv_input(arguments['--input'])
        t_drop = float(arguments['--drop'])

        print('Length of provided sequence: {} amino acids.'.format(len(wt_sequence)))
        df = drop_rows(csv_file, amino_acids, t_drop)
        no_rnd = arguments['--nornd']

        single_variants, single_values, higher_variants, higher_values = get_variants(df, amino_acids, wt_sequence)
        print('Number of single variants: {}.'.format(len(single_variants)))
        if len(single_variants) == 0:
            print('Found NO single substitution variants for possible recombination!')
        sub_ls, val_ls, sub_vs, val_vs = make_sub_ls_vs(single_variants, single_values, higher_variants, higher_values)
        print('Tip: You can edit your LS and VS datasets just by cutting/pasting between the LS and VS fasta datasets.')

        print('Creating LS dataset...', end='\r')
        make_fasta_ls_vs('LS.fasta', wt_sequence, sub_ls, val_ls)
        print('Creating VS dataset...', end='\r')
        make_fasta_ls_vs('VS.fasta', wt_sequence, sub_vs, val_vs)

        try:
            no_rnd = int(no_rnd)
        except ValueError:
            no_rnd = 0
        if no_rnd != 0:
            random_set_counter = 1
            no_rnd = int(no_rnd)
            while random_set_counter <= no_rnd:
                print('Creating random LV and VS No. {}...'.format(random_set_counter), end='\r')
                sub_ls, val_ls, sub_vs, val_vs = make_sub_ls_vs_randomly(
                    single_variants, single_values,
                    higher_variants, higher_values
                )
                make_fasta_ls_vs('LS_random_' + str(random_set_counter) + '.fasta', wt_sequence, sub_ls, val_ls)
                make_fasta_ls_vs('VS_random_' + str(random_set_counter) + '.fasta', wt_sequence, sub_vs, val_vs)
                random_set_counter += 1
        print('\n\nDone!\n')

    elif arguments['mkps']:
        wt_sequence = get_wt_sequence(arguments['--wtseq'])
        csv_file = csv_input(arguments['--input'])
        t_drop = float(arguments['--drop'])
        df = drop_rows(csv_file, amino_acids, t_drop)
        drop_wt = []
        for i in range(len(df)):
            if df.iloc[i, 0] == 'WT':
                print('Dropping wild-type (WT) from DataFrame as it cannot be used for (re-)combination.')
                drop_wt.append(i)
        df = df.drop(drop_wt).reset_index(drop=True)

        print('Length of provided sequence: {} amino acids.'.format(len(wt_sequence)))
        single_variants, _, higher_variants, _ = get_variants(df, amino_acids, wt_sequence)
        print('Using single substitution variants for (re-)combination. '
              'Number of single variants: {}.'.format(len(single_variants)))
        no_done = False
        if len(single_variants) == 0:
            print('Found NO single substitution variants for possible recombination! '
                  'No prediction files can be created!')
            no_done = True

        if arguments['--drecomb']:
            print('Creating Recomb_Double_Split...')
            for no, files in enumerate(make_combinations_double(single_variants)):
                double_mutants = np.array(files)
                create_split_files(double_mutants, single_variants, wt_sequence, 'Recomb_Double', no)

        if arguments['--trecomb']:
            print('Creating Recomb_Triple_Split...')
            for no, files in enumerate(make_combinations_triple(single_variants)):
                triple_mutants = np.array(files)
                create_split_files(triple_mutants, single_variants, wt_sequence, 'Recomb_Triple', no)

        if arguments['--qrecomb']:
            print('Beware that this step might require much disk space as PyPEF is '
                  'creating prediction files in TXT format. Creating Recomb_Quadruple_Split...')
            for no, files in enumerate(make_combinations_quadruple(single_variants)):
                quadruple_mutants = np.array(files)
                create_split_files(quadruple_mutants, single_variants, wt_sequence, 'Recomb_Quadruple', no)

        if arguments['--ddiverse']:
            print('Creating Diverse_Double_Split...')
            for no, files in enumerate(make_combinations_double_all_diverse(single_variants, amino_acids)):
                doubles = np.array(files)
                create_split_files(doubles, single_variants, wt_sequence, 'Diverse_Double', no + 1)

        if arguments['--tdiverse']:
            print('Beware that this step might require much disk space as PyPEF is '
                  'creating prediction files in TXT format. Creating Diverse_Triple_Split... ')
            for no, files in enumerate(make_combinations_triple_all_diverse(single_variants, amino_acids)):
                triples = np.array(files)
                create_split_files(triples, single_variants, wt_sequence, 'Diverse_Triple', no + 1)

        if arguments['--qdiverse']:
            print('Beware that this step might require much disk space as PyPEF is '
                  'creating prediction files in TXT format. Creating Diverse_Quadruple_Split...')
            for no, files in enumerate(make_combinations_quadruple_all_diverse(single_variants, amino_acids)):
                quadruples = np.array(files)
                create_split_files(quadruples, single_variants, wt_sequence, 'Diverse_Quadruple', no + 1)

        if arguments['--drecomb'] is False and arguments['--trecomb'] is False \
                and arguments['--qrecomb'] is False and arguments['--ddiverse'] is False \
                and arguments['--tdiverse'] is False and arguments['--qdiverse'] is False:
            print('\nInput Error:\nAt least one specification needed: Specify recombinations for mkps ; '
                  'e.g. try: "run_pypef.py mkps --drecomb" for performing double recombinant Prediction set.\n')
            no_done = True

        if no_done is False:
            print('\nDone!\n')

    elif arguments['run']:
        if arguments['--ls'] is not None and arguments['--vs'] is not None:
            if arguments['--model'] is None and arguments['--figure'] is None:
                path = os.getcwd()
                try:
                    t_save = int(arguments['--save'])
                except ValueError:
                    t_save = 5
                if arguments['--parallel']:
                    # Parallelization of AAindex iteration
                    # import parallel modules here as ray is yet not supported for Windows
                    import ray
                    ray.init()
                    from pypef.Modules_Parallelization import r2_list_parallel
                    cores = arguments['--cores']
                    try:
                        cores = int(cores)
                    except (ValueError, TypeError):
                        try:
                            cores = multiprocessing.cpu_count() // 2
                        except NotImplementedError:
                            cores = 4
                    print('Using {} cores for parallel computing. Running...'.format(cores))
                    aaindex_r2_list = r2_list_parallel(
                        arguments['--ls'], arguments['--vs'], cores,
                        arguments['--regressor'], arguments['--nofft'],
                        arguments['--sort']
                    )
                    formatted_output(aaindex_r2_list, arguments['--nofft'])
                    save_model(
                        path, aaindex_r2_list, arguments['--ls'], arguments['--vs'], t_save,
                        arguments['--regressor'], arguments['--nofft'], arguments['--all']
                    )

                else:
                    aaindex_r2_list = r2_list(
                        arguments['--ls'], arguments['--vs'], arguments['--regressor'],
                        arguments['--nofft'], arguments['--sort']
                    )
                    formatted_output(aaindex_r2_list, arguments['--nofft'])
                    save_model(
                        path, aaindex_r2_list, arguments['--ls'], arguments['--vs'], t_save,
                        arguments['--regressor'], arguments['--nofft'], arguments['--all']
                    )
                print('\nDone!\n')

        elif arguments['--figure'] is not None and arguments['--model'] is not None:
            path = os.getcwd()
            plot(
                path, arguments['--figure'], arguments['--model'], arguments['--label'],
                arguments['--color'], arguments['--ywt'], arguments['--nofft']
            )
            print('\nCreated plot!\n')

        # Prediction of single .fasta file
        elif arguments['--ps'] is not None and arguments['--model'] is not None:
            path = os.getcwd()
            predictions = predict(path, arguments['--ps'], arguments['--model'], None, arguments['--nofft'])
            if arguments['--negative']:
                predictions = sorted(predictions, key=lambda x: x[0], reverse=False)
            predictions_out(predictions, arguments['--model'], arguments['--ps'])
            print('\nDone!\n')

        # Prediction on recombinant/diverse variant folder data
        elif arguments['--pmult'] and arguments['--model'] is not None:
            path = os.getcwd()
            recombs_total = []
            recomb_d, recomb_t, recomb_q = '/Recomb_Double_Split/', '/Recomb_Triple_Split/', '/Recomb_Quadruple_Split/'
            diverse_d, diverse_t, diverse_q = '/Diverse_Double_Split/', '/Diverse_Triple_Split/', \
                                              '/Diverse_Quadruple_Split/'
            if arguments['--drecomb']:
                recombs_total.append(recomb_d)
            if arguments['--trecomb']:
                recombs_total.append(recomb_t)
            if arguments['--qrecomb']:
                recombs_total.append(recomb_q)
            if arguments['--ddiverse']:
                recombs_total.append(diverse_d)
            if arguments['--tdiverse']:
                recombs_total.append(diverse_t)
            if arguments['--qdiverse']:
                recombs_total.append(diverse_q)
            if arguments['--drecomb'] is False and arguments['--trecomb'] is False and arguments['--qrecomb'] is False \
                    and arguments['--ddiverse'] is False and arguments['--tdiverse'] is False \
                    and arguments['--qdiverse'] is False:
                print('Define prediction target for --pmult, e.g. --pmult --drecomb.')

            for args in recombs_total:
                predictions_total = []
                print('Running predictions for files in {}...'.format(args[1:-1]))
                path_recomb = path + args
                os.chdir(path)
                files = [f for f in listdir(path_recomb) if isfile(join(path_recomb, f)) if f.endswith('.fasta')]
                for f in tqdm(files):
                    predictions = predict(path, f, arguments['--model'], path_recomb, arguments['--nofft'])
                    for pred in predictions:
                        predictions_total.append(pred)  # perhaps implement numpy.save if array gets too large byte size
                predictions_total = list(dict.fromkeys(predictions_total))  # removing duplicates from list
                if arguments['--negative']:
                    predictions_total = sorted(predictions_total, key=lambda x: x[0], reverse=False)

                else:
                    predictions_total = sorted(predictions_total, key=lambda x: x[0], reverse=True)

                predictions_out(predictions_total, arguments['--model'], 'Top' + args[1:-1])
                os.chdir(path)
            print('\nDone!\n')

    # Metropolis-Hastings-driven directed evolution, similar to Biswas et al.:
    # Low-N protein engineering with data-efficient deep learning,
    # see https://github.com/ivanjayapurna/low-n-protein-engineering/tree/master/directed-evo
    elif arguments['directevo']:
        if arguments['--model'] is not None:
            print('Not counting WT as variant in directed evolution as it cannot be used for (re-)combination.')
            path = os.getcwd()
            try:
                # "temperature" parameter: determines sensitivity of Metropolis-Hastings acceptance criteria
                temp = float(arguments['--temp'])
                # how many subsequent mutation trials per simulated evolution trajectory
                num_iterations = int(arguments['--numiter'])
                # how many separate evolution trajectories to run
                num_trajectories = int(arguments['--numtraj'])
            except ValueError:
                raise ValueError("Define flags 'numiter' and 'numtraj' as integer and 'temp' as float.")

            args_model = arguments['--model']
            s_wt = get_wt_sequence(arguments['--wtseq'])
            y_wt = arguments['--ywt']
            negative = arguments['--negative']

            # Metropolis-Hastings-driven directed evolution on single mutant position csv data
            usecsv = arguments['--usecsv']
            if usecsv is True:
                csv_file = csv_input(arguments['--input'])
                t_drop = float(arguments['--drop'])

                print('Length of provided sequence: {} amino acids.'.format(len(s_wt)))
                df = drop_rows(csv_file, amino_acids, t_drop)
                drop_wt = []
                for i in range(len(df)):
                    if df.iloc[i, 0] == 'WT':
                        print('Using fitness value (y_WT) for wild-type (WT) as specified in CSV.')
                        drop_wt.append(i)
                        y_wt = df.iloc[i, 1]
                df = df.drop(drop_wt).reset_index(drop=True)

                single_variants, single_values, higher_variants, higher_values = get_variants(df, amino_acids, s_wt)
                print('Number of single variants: {}.'.format(len(single_variants)))
                if len(single_variants) == 0:
                    print('Found NO single substitution variants for possible recombination!')
                sub_ls, val_ls, _, _ = make_sub_ls_vs(
                    single_variants, single_values, higher_variants,
                    higher_values, directed_evolution=True
                )
                print('Creating single variant dataset...')

                make_fasta_ls_vs('Single_variants.fasta', s_wt, sub_ls, val_ls)

            else:
                sub_ls = None

            # Metropolis-Hastings-driven directed evolution on single mutant .csv amino acid substitution data
            csvaa = arguments['--csvaa']
            traj_records_folder = 'DE_record'

            print('Running evolution trajectories and plotting...')

            run_de_trajectories(
                s_wt, args_model, y_wt, num_iterations, num_trajectories,
                traj_records_folder, amino_acids, temp, path, sub_ls, arguments['--nofft'],
                negative=negative, save=True, usecsv=usecsv, csvaa=csvaa,
                print_matrix=arguments['--print']
            )
            print('\nDone!')


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for making prediction files
from input CSV file (with value separators sep=',' or sep=';')
having the CSV-format

HEADER_VARIANTS;HEADER_VARIANTS_FITNESS
VARIANT_1;FITNESS_VALUE_1
VARIANT_2;FITNESS_VALUE_2
...

according to the self devised prediction set convention
> VARIANT_NAME_1
VARIANT_SEQUENCE_1
> VARIANT_NAME_2
VARIANT_SEQUENCE_2
...
"""


import os
import numpy as np
from tqdm import tqdm


def make_fasta_ps(filename, wt, substitution):
    """
    Creates prediction sets (.fasta style files)
    """
    myfile = open(filename, 'w')
    count = 0
    for i, var in enumerate(substitution):
        temporary = list(wt)
        name = ''
        separation = 0
        for single_var in var:
            position_index = int(str(single_var)[1:-1]) - 1
            new_amino_acid = str(single_var)[-1]
            temporary[position_index] = new_amino_acid
            if separation == 0:
                name += single_var
            else:
                name += '/' + single_var
            separation += 1
        print('>', name, file=myfile)
        print(''.join(temporary), file=myfile)
        count += 1
    myfile.close()


def make_combinations_double(arr):
    """
    Make double recombination variants
    """
    doubles = []
    for i in tqdm(range(len(arr))):
        for j in range(len(arr)):
            if j > i:
                if (arr[i][0])[1:-1] != (arr[j][0])[1:-1]:
                    doubles.append([arr[i][0], arr[j][0]])
                    if len(doubles) >= 8E04:
                        yield doubles
                        doubles = []
    yield doubles


def make_combinations_triple(arr):
    """
    Make triple recombination variants
    """
    length = len(arr)
    triples = []
    for i in tqdm(range(length)):
        for j in range(length):
            for k in range(length):
                if k > j > i:
                    if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] != (arr[k][0])[1:-1]:
                        triples.append([arr[i][0], arr[j][0], arr[k][0]])
                        if len(triples) >= 8E04:
                            yield triples
                            triples = []
    yield triples


def make_combinations_quadruple(arr):
    """
    Make quadruple recombination variants
    """
    length = len(arr)
    quadruples = []
    for i in tqdm(range(length)):
        for j in range(length):
            for k in range(length):
                for l in range(length):
                    if l > k > j > i:
                        if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] != (arr[k][0])[1:-1] != (arr[l][0])[1:-1]:
                            quadruples.append([arr[i][0], arr[j][0], arr[k][0], arr[l][0]])
                            if len(quadruples) >= 8E04:
                                yield quadruples
                                quadruples = []
    yield quadruples


def make_directory_and_enter(directory):
    """
    Makes directory for recombined or diverse prediction sets
    """
    previous_working_directory = os.getcwd()
    try:
        if not os.path.exists(os.path.dirname(directory)):
            os.mkdir(directory)
    except OSError:
        pass
    os.chdir(directory)

    return previous_working_directory


def create_split_files(array, single_variants, wt_sequence, name, no):
    """
    Creates split files from given variants for yielded recombined or diverse variants
    """
    if len(array) > 0:
        number_of_split_files = len(array) / (len(single_variants) * 20 ** 3)
        number_of_split_files = round(number_of_split_files)
        if number_of_split_files == 0:
            number_of_split_files += 1
        split = np.array_split(array, number_of_split_files)
        pwd = make_directory_and_enter(name + '_Split')
        for i in split:
            name_ = name + '_Split' + str(no) + '.fasta'
            make_fasta_ps(name_, wt_sequence, i)

        os.chdir(pwd)

        return ()


def make_combinations_double_all_diverse(arr, aminoacids):
    """
    Make double substituted naturally diverse variants
    """
    doubles = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in aminoacids:
                for l in aminoacids:
                    if ((arr[i][0])[1:-1]) != ((arr[j][0])[1:-1]) and\
                            ((arr[i][0])[:-1] + k)[0] != ((arr[i][0])[:-1] + k)[-1] and\
                            ((arr[j][0])[:-1] + l)[0] != ((arr[j][0])[:-1] + l)[-1]:
                        doubles.append(tuple([(arr[i][0])[:-1] + k, (arr[j][0])[:-1] + l]))  # tuple needed for
                        if len(doubles) >= 8E04:                                             # list(dict()):
                            doubles = list(dict.fromkeys(doubles))  # removes duplicated list entries
                            yield doubles
                            doubles = []
    doubles = list(dict.fromkeys(doubles))
    yield doubles


def make_combinations_triple_all_diverse(arr, aminoacids):
    """
    Make triple substituted naturally diverse variants
    """
    triples = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in range(j + 1, len(arr)):
                for l in aminoacids:
                    for m in aminoacids:
                        for n in aminoacids:
                            if ((arr[i][0])[1:-1]) != ((arr[j][0])[1:-1]) != ((arr[k][0])[1:-1]) and\
                                    ((arr[i][0])[:-1] + l)[0] != ((arr[i][0])[:-1] + l)[-1] and\
                                    ((arr[j][0])[:-1] + m)[0] != ((arr[j][0])[:-1] + m)[-1] and\
                                    ((arr[k][0])[:-1] + n)[0] != ((arr[k][0])[:-1] + n)[-1]:
                                triples.append(tuple([(arr[i][0])[:-1] + l, (arr[j][0])[:-1] + m,
                                                      (arr[k][0])[:-1] + n]))
                                if len(triples) >= 8E04:
                                    triples = list(dict.fromkeys(triples))  # transfer to dict and back to list
                                    yield triples
                                    triples = []
    triples = list(dict.fromkeys(triples))
    yield triples


def make_combinations_quadruple_all_diverse(arr, aminoacids):
    """
    Make quadruple substituted naturally diverse variants
    """
    quadruples = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in range(j + 1, len(arr)):
                for l in range(k + 1, len(arr)):
                    for m in aminoacids:
                        for n in aminoacids:
                            for o in aminoacids:
                                for p in aminoacids:
                                    if ((arr[i][0])[1:-1]) != ((arr[j][0])[1:-1]) != ((arr[k][0])[1:-1]) != \
                                            ((arr[l][0])[1:-1]) and\
                                            ((arr[i][0])[:-1] + m)[0] != ((arr[i][0])[:-1] + m)[-1] and\
                                            ((arr[j][0])[:-1] + n)[0] != ((arr[j][0])[:-1] + n)[-1] and\
                                            ((arr[k][0])[:-1] + o)[0] != ((arr[k][0])[:-1] + o)[-1] and\
                                            ((arr[l][0])[:-1] + p)[0] != ((arr[l][0])[:-1] + p)[-1]:
                                        quadruples.append(tuple([(arr[i][0])[:-1] + m, (arr[j][0])[:-1] + n,
                                                                 (arr[k][0])[:-1] + o, (arr[l][0])[:-1] + p]))
                                        if len(quadruples) >= 8E04:
                                            quadruples = list(dict.fromkeys(quadruples))  # transfer to dict
                                            yield quadruples                              # and back to list
                                            quadruples = []
    quadruples = list(dict.fromkeys(quadruples))
    yield quadruples



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Main modules for regression/ML including featurization,
validation, tuning, prediction, and plotting routines.
"""

import os
import sys
import matplotlib
matplotlib.use('Agg')  # no plt.show(), just save plot
import matplotlib.pyplot as plt
import numpy as np
import pickle
from tqdm import tqdm  # progress bars
# import adjust_text  # only locally imported for labeled validation plots and in silico directed evolution
from sklearn.model_selection import LeaveOneOut
from sklearn.model_selection import KFold
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import GridSearchCV  # default: refit=True

# import regression models
from sklearn.cross_decomposition import PLSRegression
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor

import warnings
# ignoring warnings of PLS regression using n_components
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')


def read_models(number):
    """
    reads the models found in the file Model_Results.txt.
    If no model was trained, the .txt file does not exist.
    """
    try:
        ls = ""
        with open('Model_Results.txt', 'r') as file:
            for i, lines in enumerate(file):
                if i == 0:
                    if lines[:6] == 'No FFT':
                        number += 2
                if i <= number + 1:
                    ls += lines
        return ls
    except FileNotFoundError:
        return "No Model_Results.txt found."


def full_path(filename):
    """
    returns the path of an index inside the folder /AAindex/,
    e.g. path/to/AAindex/FAUJ880109.txt.
    """
    modules_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(modules_path, 'AAindex/' + filename)


def path_aaindex_dir():
    """
    returns the path to the /AAindex folder, e.g. path/to/AAindex/.
    """
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'AAindex')


class AAIndex:
    """
    gets all the information that are given in each AAindex file.
    For the program routine it provides the library to enable translation
    of the alphabetical amino acid sequence to an array of numericals.
    """
    def __init__(self, filename):
        self.File = filename
        self.Accession_Number = None
        self.Data_Description = None
        self.PMID = None
        self.Authors = None
        self.Title_Of_Article = None
        self.Journal_Reference = None

    def general_information(self):
        """
        Gets and allocates general information based on the AAindex file
        format defined by file sections 'H', 'D', 'E', 'A', 'T', 'J'
        """
        with open(self.File, 'r') as f:
            for line in f:
                # try/ except "removes" empty lines.
                try:
                    words = line.split()
                    id_letter = words[0]
                except IndexError:
                    break

                # Extract some general information about AAindex file.
                if id_letter == 'H':
                    self.Accession_Number = words[1]
                elif id_letter == 'D':
                    self.Data_Description = words[1]
                elif id_letter == 'E':
                    self.PMID = words[1:]
                elif id_letter == 'A':
                    self.Authors = ' '.join(words[1:])
                elif id_letter == 'T':
                    self.Title_Of_Article = ' '.join(words[1:])
                elif id_letter == 'J':
                    self.Journal_Reference = ' '.join(words[1:])

    def encoding_dictionary(self):
        """
        Get numerical values of AAindex for each amino acid
        """
        with open(self.File, 'r') as f:
            for line in f:
                # try/ except "removes" empty lines
                try:
                    words = line.split()
                    id_letter = words[0]
                except IndexError:
                    break

                # Extract numerical values of AAindex.
                if id_letter == 'I':

                    keys = []
                    for word in words[1:]:
                        keys.append(word[0])
                        keys.append(word[-1])

                    values = []
                    for row in range(2):
                        line = f.readline()
                        strings = line.split()
                        for idx, string in enumerate(strings):
                            # Some amino acids may have no value
                            try:
                                strings[idx] = float(string)
                            except ValueError:
                                strings[idx] = None
                        values.append(strings)
                    values = np.reshape(np.array(values).T, len(keys))

                    return dict(zip(keys, values))


def get_sequences(fasta, mult_path=None, prediction=False):
    """
    "Get_Sequences" reads (learning and validation).fasta format files and extracts the name,
    the target value and the sequence of the peptide. See example directory for required fasta file format.
    Make sure every marker (> and ;) is seperated by an space ' ' from the value/ name.
    """
    if mult_path is not None:
        os.chdir(mult_path)

    sequences = []
    values = []
    names_of_mutations = []

    with open(fasta, 'r') as f:
        for line in f:
            if '>' in line:
                words = line.split()
                names_of_mutations.append(words[1])
                # words[1] is appended so make sure there is a space in between > and the name!

            elif '#' in line:
                pass  # are Comments

            elif ';' in line:
                words = line.split()
                values.append(float(words[1]))
                # words[1] is appended so make sure there is a space in between ; and the value!

            else:
                try:
                    words = line.split()
                    sequences.append(words[0])
                except IndexError:
                    raise IndexError("Learning or Validation sets (.fasta) likely have emtpy lines at end of file")

    # Check consistency
    if not prediction:
        if len(sequences) != len(values):
            print('Error: Number of sequences does not fit with number of target values!')
            print('Number of sequences: {}, Number of target values: {}.'.format(str(len(sequences)), str(len(values))))
            sys.exit()

    return sequences, names_of_mutations, values


class XY:
    """
    converts the string sequence into a list of numericals using the AAindex translation library,
    Fourier transforming the numerical array that was translated by get_numerical_sequence --> do_fourier,
    computing the input matrices X and Y for the PLS regressor (get_x_and_y). Returns FFT-ed arrays (x),
    labels array (y), and raw_encoded sequences arrays (raw_numerical_sequences)
    """
    def __init__(self, aaindex_file, fasta_file, mult_path=None, prediction=False):
        aaidx = AAIndex(aaindex_file)
        self.dictionary = aaidx.encoding_dictionary()
        self.sequences, self.names, self.values = get_sequences(fasta_file, mult_path, prediction)

    def get_numerical_sequence(self, sequence):
        return np.array([self.dictionary[aminoacid] for aminoacid in sequence])

    @staticmethod
    def do_fourier(sequence):
        """
        This static function does the Fast Fourier Transform. Since the condition

                    len(Array) = 2^k -> k = log_2(len(Array))
                    k in N

        must be satisfied, the array must be reshaped (zero padding) if k is no integer value.
        The verbose parameter prints also the real and imaginary part separately.
        """
        threshold = 1e-8  # errors due to computer uncertainties
        k = np.log2(sequence.size)  # get exponent k
        mean = np.mean(sequence, axis=0)  # calculate mean of numerical array
        sequence = np.subtract(sequence, mean)  # subtract mean to avoid artificial effects of FT

        if abs(int(k) - k) > threshold:  # check if length of array fulfills previous equation
            numerical_sequence_reshaped = np.zeros(pow(2, (int(k) + 1)))  # reshape array
            for index, value in enumerate(sequence):
                numerical_sequence_reshaped[index] = value
            sequence = numerical_sequence_reshaped

        fourier_transformed = np.fft.fft(sequence)  # FFT
        ft_real = np.real(fourier_transformed)
        ft_imag = np.imag(fourier_transformed)

        x = np.linspace(1, sequence.size, sequence.size)  # frequencies
        x = x / max(x)  # normalization of frequency

        amplitude = ft_real * ft_real + ft_imag * ft_imag

        if max(amplitude) != 0:
            amplitude = np.true_divide(amplitude, max(amplitude))  # normalization of amplitude

        return amplitude, x

    def get_x_and_y(self):
        """
        getting the input matrices X (FFT amplitudes) and Y (variant labels)
        """
        frequencies = []
        amplitudes = []
        raw_numerical_seq = []

        for sequence in self.sequences:
            num = self.get_numerical_sequence(sequence)

            # There may be amino acids without a value in AAindex.
            # Skip these Indices.
            if None in num:
                break

            # Numerical sequence gets expended by zeros so that also different lengths of sequences
            # can be processed using '--nofft' option
            k = np.log2(len(num))
            if abs(int(k) - k) > 1e-8:  # check if length of array fulfills previous equation
                num_appended = np.append(num, np.zeros(pow(2, (int(k) + 1)) - len(num)))  # reshape array
            else:
                num_appended = num

            amplitudes_, frequencies_ = self.do_fourier(num)

            # Fourier spectra are mirrored at frequency = 0.5. No more information at higher frequencies.
            half = len(frequencies_) // 2  # // for integer division
            frequencies.append(frequencies_[:half])
            amplitudes.append(amplitudes_[:half])    # FFT-ed encoded amino acid sequences
            raw_numerical_seq.append(num_appended)  # Raw encoded amino acid sequences

        amplitudes = np.array(amplitudes)
        # frequencies = np.array(frequencies)  # not used
        raw_numerical_seq = np.array(raw_numerical_seq)

        x = amplitudes
        y = self.values                             # Fitness values (sequence labels)

        return x, y, raw_numerical_seq


def get_r2(x_learn, x_valid, y_learn, y_valid, regressor='pls'):
    """
    The function Get_R2 takes features and labels from the learning and validation set.

    When using 'pls' as regressor, the MSE is calculated for all LOOCV sets for predicted vs true labels
    (mse = mean_squared_error(y_test_loo, y_pred_loo) for a fixed number of components for PLS regression.
    In the next iteration, the number of components is increased by 1 (number_of_components += 1)
    and the MSE is calculated for this regressor. The loop breaks if i > 9.
    Finally, the model of the single AAindex model with the lowest MSE is chosen.

    When using other regressors the parameters are tuned using GridSearchCV.

    This function returnes performance (R2, (N)RMSE, Pearson's r) and model parameters.
    """
    regressor = regressor.lower()
    mean_squared_error_list = []

    if regressor == 'pls':
        # PLS regression with LOOCV n_components tuning as described by Cadet et al.
        # https://doi.org/10.1186/s12859-018-2407-8
        # https://doi.org/10.1038/s41598-018-35033-y
        # Hyperparameter (N component) tuning of PLS regressor
        for n_comp in range(1, 10):  # n_comp = 1, 2,..., 9
            pls = PLSRegression(n_components=n_comp)
            loo = LeaveOneOut()

            y_pred_loo = []
            y_test_loo = []

            for train, test in loo.split(x_learn):
                x_learn_loo = []
                y_learn_loo = []
                x_test_loo = []

                for j in train:
                    x_learn_loo.append(x_learn[j])
                    y_learn_loo.append(y_learn[j])

                for k in test:
                    x_test_loo.append(x_learn[k])
                    y_test_loo.append(y_learn[k])

                pls.fit(x_learn_loo, y_learn_loo)
                y_pred_loo.append(pls.predict(x_test_loo)[0][0])

            mse = mean_squared_error(y_test_loo, y_pred_loo)

            mean_squared_error_list.append(mse)

        mean_squared_error_list = np.array(mean_squared_error_list)
        # idx = np.where(...) finds best number of components
        idx = np.where(mean_squared_error_list == np.min(mean_squared_error_list))[0][0] + 1

        # Model is fitted with best n_components (lowest MSE)
        best_params = {'n_components': idx}
        regressor_ = PLSRegression(n_components=best_params.get('n_components'))

    # other regression options (CV tuning)
    elif regressor == 'pls_cv':
        params = {'n_components': list(np.arange(1, 10))}  # n_comp = 1, 2,..., 9
        regressor_ = GridSearchCV(PLSRegression(), param_grid=params, iid=False, cv=5)  # iid in future
                                                                                        # versions redundant
    elif regressor == 'rf':
        params = {                 # similar parameter grid as Xu et al., https://doi.org/10.1021/acs.jcim.0c00073
            'random_state': [42],  # state determined
            'n_estimators': [100, 250, 500, 1000],  # number of individual decision trees in the forest
            'max_features': ['auto', 'sqrt', 'log2']  # “auto” -> max_features=n_features,
            # “sqrt” -> max_features=sqrt(n_features) “log2” -> max_features=log2(n_features)
        }
        regressor_ = GridSearchCV(RandomForestRegressor(), param_grid=params, iid=False, cv=5)

    elif regressor == 'svr':
        params = {                      # similar parameter grid as Xu et al.
            'C': [2 ** 0, 2 ** 2, 2 ** 4, 2 ** 6, 2 ** 8, 2 ** 10, 2 ** 12],  # Regularization parameter
            'gamma': [0.1, 0.01, 0.001, 0.0001, 0.00001]  # often 1 / n_features or 1 / (n_features * X.var())
        }
        regressor_ = GridSearchCV(SVR(), param_grid=params, iid=False, cv=5)

    elif regressor == 'mlp':
        params = {
            # feedforward network trained via backpropagation – here only using a single hidden layer
            'hidden_layer_sizes': [i for i in range(1, 12)],  # size of hidden layer [(1,), (2,), ..., (12,)]
            'activation': ['relu'],  # rectified linear unit
            'solver': ['adam', 'lbfgs'],  # ADAM: A Method for Stochastic Optimization , or Limited-memory BFGS
            'learning_rate': ['constant'],  # learning rate given by ‘learning_rate_init’
            'learning_rate_init': [0.001, 0.01, 0.1],  # only used when solver=’sgd’ or ‘adam’
            'max_iter': [1000, 200],  # for stochastic solvers (‘sgd’, ‘adam’) determines epochs
            'random_state': [42]
        }
        regressor_ = GridSearchCV(MLPRegressor(), param_grid=params, iid=False, cv=5)

    else:
        raise SystemError("Did not find specified regression model as valid option. See '--help' for valid "
                          "regression model options.")

    regressor_.fit(x_learn, y_learn)  # fit model

    if regressor != 'pls':      # take best parameters for the regressor and the AAindex
        best_params = regressor_.best_params_

    y_pred = []
    try:
        for y_p in regressor_.predict(x_valid):  # predict validation entries with fitted model
            y_pred.append(float(y_p))
    except ValueError:
        raise ValueError("Above error message exception indicates that your validation set may be empty.")

    r2 = r2_score(y_valid, y_pred)
    rmse = np.sqrt(mean_squared_error(y_valid, y_pred))
    nrmse = rmse / np.std(y_valid, ddof=1)
    # ranks for Spearman's rank correlation
    y_val_rank = np.array(y_valid).argsort().argsort()
    y_pred_rank = np.array(y_pred).argsort().argsort()
    with warnings.catch_warnings():  # catching RunTime warning when there's no variance in an array, e.g. [2, 2, 2, 2]
        warnings.simplefilter("ignore")  # which would mean divide by zero
        pearson_r = np.corrcoef(y_valid, y_pred)[0][1]
        spearman_rho = np.corrcoef(y_val_rank, y_pred_rank)[0][1]

    return r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params


def r2_list(learning_set, validation_set, regressor='pls', no_fft=False, sort='1'):
    """
    returns the sorted list of all the model parameters and
    the performance values (r2 etc.) from function get_r2.
    """
    aa_indices = [file for file in os.listdir(path_aaindex_dir()) if file.endswith('.txt')]

    aaindex_r2_list = []
    for index, aaindex in enumerate(tqdm(aa_indices)):
        xy_learn = XY(full_path(aaindex), learning_set)
        if not no_fft:  # X is FFT-ed of encoded alphabetical sequence
            x_learn, y_learn, _ = xy_learn.get_x_and_y()
        else:               # X is raw encoded of alphabetical sequence
            _, y_learn, x_learn = xy_learn.get_x_and_y()

        # If x_learn (or y_learn) is an empty array, the sequence could not be encoded,
        # because of NoneType value. -> Skip
        if len(x_learn) != 0:
            xy_test = XY(full_path(aaindex), validation_set)
            if not no_fft:  # X is FFT-ed of the encoded alphabetical sequence
                x_test, y_test, _ = xy_test.get_x_and_y()
            else:               # X is the raw encoded of alphabetical sequence
                _, y_test, x_test = xy_test.get_x_and_y()
            r2, rmse, nrmse, pearson_r, spearman_rho, regression_model, params = get_r2(x_learn, x_test, y_learn,
                                                                                        y_test, regressor)
            aaindex_r2_list.append([aaindex, r2, rmse, nrmse, pearson_r, spearman_rho, regression_model, params])

    try:
        sort = int(sort)
        if sort == 2 or sort == 3:
            aaindex_r2_list.sort(key=lambda x: x[sort])
        else:
            aaindex_r2_list.sort(key=lambda x: x[sort], reverse=True)

    except ValueError:
        raise ValueError("Choose between options 1 to 5 (R2, RMSE, NRMSE, Pearson's r, Spearman's rho.")

    return aaindex_r2_list


def formatted_output(aaindex_r2_list, no_fft=False, minimum_r2=0.0):
    """
    takes the sorted list from function r2_list and writes the model names with an R2 ≥ 0
    as well as the corresponding parameters for each model so that the user gets
    a list (Model_Results.txt) of the top ranking models for the given validation set.
    """

    index, value, value2, value3, value4, value5, regression_model, params = [], [], [], [], [], [], [], []

    for (idx, val, val2, val3, val4, val5, r_m, pam) in aaindex_r2_list:
        if val >= minimum_r2:
            index.append(idx[:-4])
            value.append('{:f}'.format(val))
            value2.append('{:f}'.format(val2))
            value3.append('{:f}'.format(val3))
            value4.append('{:f}'.format(val4))
            value5.append('{:f}'.format(val5))
            regression_model.append(r_m.upper())
            params.append(pam)

    if len(value) == 0:  # Criterion of not finding suitable model is defined by Minimum_R2
        raise ValueError('No model with positive R2.')

    data = np.array([index, value, value2, value3, value4, value5, regression_model, params]).T
    col_width = max(len(str(value)) for row in data for value in row[:-1]) + 5

    head = ['Index', 'R2', 'RMSE', 'NRMSE', 'Pearson\'s r', 'Spearman\'s rho', 'Regression', 'Model parameters']
    with open('Model_Results.txt', 'w') as f:
        if no_fft is not False:
            f.write("No FFT used in this model construction, performance"
                    " represents model accuracies on raw encoded sequence data.\n\n")

        heading = "".join(caption.ljust(col_width) for caption in head) + '\n'
        f.write(heading)

        row_length = []
        for row in data:
            row_ = "".join(str(value).ljust(col_width) for value in row) + '\n'
            row_length.append(len(row_))
        row_length_max = max(row_length)
        f.write(row_length_max * '-' + '\n')

        for row in data:
            f.write("".join(str(value).ljust(col_width) for value in row) + '\n')

    return ()


def cross_validation(x, y, regressor_, n_samples=5):
    # perform k-fold cross-validation on all data
    # k = Number of splits, change for changing k in k-fold split-up, default=5
    y_test_total = []
    y_predicted_total = []

    kf = KFold(n_splits=n_samples, shuffle=True)
    for train_index, test_index in kf.split(y):
        y = np.array(y)
        try:
            x_train, x_test = x[train_index], x[test_index]
            y_train, y_test = y[train_index], y[test_index]

            for numbers in y_test:
                y_test_total.append(numbers)
            regressor_.fit(x_train, y_train)  # Fitting on a random subset for Final_Model
            # (and on a subset subset for Learning_Model)
            # Predictions for samples in the test_set during that iteration
            y_pred_test = regressor_.predict(x_test)

            for values in y_pred_test:
                y_predicted_total.append(float(values))
        except UserWarning:
            continue

    return y_test_total, y_predicted_total


def save_model(path, aaindex_r2_list, learning_set, validation_set, threshold=5, regressor='pls',
               no_fft=False, train_on_all=False):
    """
    Function Save_Model saves the best -s THRESHOLD models as 'Pickle' files (pickle.dump),
    which can be loaded again for doing predictions. Also, in Save_Model included is the def cross_validation
    -based computing of the k-fold CV performance of the n component-optimized model on all data
    (learning + validation set); by default  k  is 5 (n_samples = 5).
    Plots of the CV performance for the t best models are stored inside the folder CV_performance.
    """
    regressor = regressor.lower()
    try:
        os.mkdir('CV_performance')
    except FileExistsError:
        pass
    try:
        os.mkdir('Pickles')
    except FileExistsError:
        pass

    try:
        os.remove('CV_performance/_CV_Results.txt')
    except FileNotFoundError:
        pass
    file = open('CV_performance/_CV_Results.txt', 'w')
    file.write('5-fold cross-validated performance of top models for validation set across all data.\n\n')
    if no_fft:
        file.write("No FFT used in this model construction, performance represents"
                   " model accuracies on raw encoded sequence data.\n\n")
    file.close()

    for t in range(threshold):
        try:
            idx = aaindex_r2_list[t][0]
            parameter = aaindex_r2_list[t][7]

            # Estimating the CV performance of the n_component-fitted model on all data
            xy_learn = XY(full_path(idx), learning_set)
            xy_test = XY(full_path(idx), validation_set)
            if no_fft is False:
                x_test, y_test, _ = xy_test.get_x_and_y()
                x_learn, y_learn, _ = xy_learn.get_x_and_y()
            else:
                _, y_test, x_test = xy_test.get_x_and_y()
                _, y_learn, x_learn = xy_learn.get_x_and_y()

            x = np.concatenate([x_learn, x_test])
            y = np.concatenate([y_learn, y_test])

            if regressor == 'pls' or regressor == 'pls_cv':
                # n_components according to lowest MSE for validation set
                regressor_ = PLSRegression(n_components=parameter.get('n_components'))

            elif regressor == 'rf':
                regressor_ = RandomForestRegressor(
                    random_state=parameter.get('random_state'),
                    n_estimators=parameter.get('n_estimators'),
                    max_features=parameter.get('max_features')
                )

            elif regressor == 'svr':
                regressor_ = SVR(C=parameter.get('C'), gamma=parameter.get('gamma'))

            elif regressor == 'mlp':
                regressor_ = MLPRegressor(
                    hidden_layer_sizes=parameter.get('hidden_layer_sizes'),
                    activation=parameter.get('activation'),
                    solver=parameter.get('solver'),
                    learning_rate=parameter.get('learning_rate'),
                    learning_rate_init=parameter.get('learning_rate_init'),
                    max_iter=parameter.get('max_iter'),
                    random_state=parameter.get('random_state')
                )

            else:
                raise SystemError("Did not find specified regression model as valid option. "
                                  "See '--help' for valid regression model options.")

            # perform 5-fold cross-validation on all data (on X and Y)
            n_samples = 5
            y_test_total, y_predicted_total = cross_validation(x, y, regressor_, n_samples)

            r_squared = r2_score(y_test_total, y_predicted_total)
            rmse = np.sqrt(mean_squared_error(y_test_total, y_predicted_total))
            stddev = np.std(y_test_total, ddof=1)
            nrmse = rmse / stddev
            pearson_r = np.corrcoef(y_test_total, y_predicted_total)[0][1]
            # ranks for Spearman correlation
            y_test_total_rank = np.array(y_test_total).argsort().argsort()
            y_predicted_total_rank = np.array(y_predicted_total).argsort().argsort()
            spearman_rho = np.corrcoef(y_test_total_rank, y_predicted_total_rank)[0][1]

            with open('CV_performance/_CV_Results.txt', 'a') as f:
                f.write('Regression type: {}; Parameter: {}; Encoding index: {}\n'.format(
                    regressor.upper(), parameter, idx[:-4]))
                f.write('R2 = {:.5f}; RMSE = {:.5f}; NRMSE = {:.5f}; Pearson\'s r = {:.5f};'
                        ' Spearman\'s rho = {:.5f}\n\n'.format(r_squared, rmse, nrmse, pearson_r, spearman_rho))

            figure, ax = plt.subplots()
            ax.scatter(y_test_total, y_predicted_total, marker='o', s=20, linewidths=0.5, edgecolor='black')
            ax.plot([min(y_test_total) - 1, max(y_test_total) + 1],
                    [min(y_predicted_total) - 1, max(y_predicted_total) + 1], 'k', lw=2)
            ax.legend([
                '$R^2$ = {}\nRMSE = {}\nNRMSE = {}\nPearson\'s $r$ = {}\nSpearman\'s '.format(
                    round(r_squared, 3), round(rmse, 3), round(nrmse, 3), round(pearson_r, 3))
                + r'$\rho$ = {}'.format(str(round(spearman_rho, 3)))
            ])
            ax.set_xlabel('Measured')
            ax.set_ylabel('Predicted')
            plt.savefig('CV_performance/' + idx[:-4] + '_' + str(n_samples) + '-fold-CV.png', dpi=250)
            plt.close('all')

            if train_on_all:
                # fit on all available data (learning + validation set; FFT or noFFT is defined already above)
                regressor_.fit(x, y)
            else:
                # fit (only) on full learning set (FFT or noFFT is defined already above)
                regressor_.fit(x_learn, y_learn)

            file = open(os.path.join(path, 'Pickles/'+idx[:-4]), 'wb')
            pickle.dump(regressor_, file)
            file.close()

        except IndexError:
            break

    return ()


def predict(path, prediction_set, model, mult_path=None, no_fft=False, print_matrix=False):
    """
    The function Predict is used to perform predictions.
    Saved pickle files of models will be loaded again (mod = pickle.load(file))
    and used for predicting the label y (y = mod.predict(x)) of sequences given in the Prediction_Set.fasta.
    """
    aaidx = full_path(str(model) + '.txt')
    xy = XY(aaidx, prediction_set, mult_path, prediction=True)
    x, _, x_raw = xy.get_x_and_y()

    file = open(os.path.join(path, 'Pickles/'+str(model)), 'rb')
    mod = pickle.load(file)
    file.close()

    try:
        y_list = []
        if no_fft is False:
            y = mod.predict(x)
            for y_ in y:
                # just make sure predicted Y is nested (list of list) [[Y_1], [Y_2], ..., [Y_N]]
                y_list.append([float(y_)])
        else:
            y = mod.predict(x_raw)
            for y_ in y:
                # just make sure predicted Y is nested (list of list) [[Y_1], [Y_2], ..., [Y_N]]
                y_list.append([float(y_)])

    except ValueError:
        raise ValueError("You likely tried to predict using a model with (or without) FFT featurization ('--nofft')"
                         " while the model was trained without (or with) FFT featurization. Check the Model_Results.txt"
                         " line 1, if the models were trained using FFT.")

    _, names_of_mutations, _ = get_sequences(prediction_set, mult_path, prediction=True)

    predictions = [(y_list[i][0], names_of_mutations[i]) for i in range(len(y_list))]

    # Pay attention if more negative values would define a better variant --> use negative flag
    predictions.sort()
    predictions.reverse()
    # if predictions array too large?  if Mult_Path is not None: predictions = predictions[:100000]

    # Print FFT-ed and raw sequence vectors for directed evolution if desired
    if print_matrix:
        print('X (FFT):\n{} len(X_raw): {}\nX_raw (noFFT):\n{} len(X): {}\n(Predicted value, Variant): {}\n\n'.format(
            x, len(x[0]), x_raw, len(x_raw[0]), predictions))

    return predictions


def predictions_out(predictions, model, prediction_set):
    """
    Writes predictions (of the new sequence space) to text file(s).
    """
    name, value = [], []
    for (val, nam) in predictions:
        name.append(nam)
        value.append('{:f}'.format(val))

    data = np.array([name, value]).T
    col_width = max(len(str(value)) for row in data for value in row) + 5

    head = ['Name', 'Prediction']
    with open('Predictions_' + str(model) + '_' + str(prediction_set)[:-6] + '.txt', 'w') as f:
        f.write("".join(caption.ljust(col_width) for caption in head) + '\n')
        f.write(len(head)*col_width*'-' + '\n')
        for row in data:
            f.write("".join(str(value).ljust(col_width) for value in row) + '\n')

    return ()


def plot(path, fasta_file, model, label, color, y_wt, no_fft=False):
    """
    Function Plot is used to make plots of the validation process and
    shows predicted (Y_pred) vs. measured/"true" (Y_true) protein fitness and
    calculates the corresponding model performance (R2, (N)RMSE, Pearson's r).
    Also allows colored version plotting to classify predictions in true or
    false positive or negative predictions.
    """
    aaidx = full_path(str(model) + '.txt')
    xy = XY(aaidx, fasta_file, prediction=False)
    x, y_true, x_raw = xy.get_x_and_y()

    try:
        file = open(os.path.join(path, 'Pickles/'+str(model)), 'rb')
        mod = pickle.load(file)
        file.close()

        y_pred = []

        try:
            # predicting (again) with (re-)loaded model (that was trained on training or full data)
            if no_fft is False:
                y_pred_ = mod.predict(x)
            else:
                y_pred_ = mod.predict(x_raw)
        except ValueError:
            raise ValueError("You likely tried to plot a validation set with (or without) FFT featurization ('--nofft')"
                             " while the model was trained without (or with) FFT featurization. Check the"
                             " Model_Results.txt line 1, if the models were trained using FFT.")

        for y_p in y_pred_:
            y_pred.append(float(y_p))

        _, names_of_mutations, _ = get_sequences(fasta_file)

        r_squared = r2_score(y_true, y_pred)
        rmse = np.sqrt(mean_squared_error(y_true, y_pred))
        stddev = np.std(y_true, ddof=1)
        nrmse = rmse / stddev
        pearson_r = np.corrcoef(y_true, y_pred)[0][1]
        # ranks for Spearman
        y_true_total_rank = np.array(y_true).argsort().argsort()
        y_pred_total_rank = np.array(y_pred).argsort().argsort()
        spearman_rho = np.corrcoef(y_true_total_rank, y_pred_total_rank)[0][1]
        legend = '$R^2$ = {}\nRMSE = {}\nNRMSE = {}\nPearson\'s $r$ = {}\nSpearman\'s '.format(
            round(r_squared, 3), round(rmse, 3), round(nrmse, 3), round(pearson_r, 3)) \
            + r'$\rho$ = {}'.format(round(spearman_rho, 3))
        x = np.linspace(min(y_pred) - 1, max(y_pred) + 1, 100)

        fig, ax = plt.subplots()
        ax.scatter(y_true, y_pred, label=legend, marker='o', s=20, linewidths=0.5, edgecolor='black')
        ax.plot(x, x, color='black', linewidth=0.5)  # plot diagonal line

        if label is not False:
            from adjustText import adjust_text
            texts = [ax.text(y_true[i], y_pred[i], txt, fontsize=4) for i, txt in enumerate(names_of_mutations)]
            adjust_text(texts, only_move={'points': 'y', 'text': 'y'}, force_points=0.5)

        if color is not False:
            try:
                y_wt = float(y_wt)
            except TypeError:
                raise TypeError('Needs label value of WT (y_WT) when making color plot (e.g. --color --ywt 1.0)')
            if y_wt == 0:
                y_wt = 1E-9  # choose a value close to zero
            true_v, true_p, false_v, false_p = [], [], [], []
            for i, v in enumerate(y_true):
                if y_true[i] / y_wt >= 1 and y_pred[i] / y_wt >= 1:
                    true_v.append(y_true[i]), true_p.append(float(y_pred[i]))
                elif y_true[i] / y_wt < 1 and y_pred[i] / y_wt < 1:
                    true_v.append(y_true[i]), true_p.append(float(y_pred[i]))
                else:
                    false_v.append(y_true[i]), false_p.append(float(y_pred[i]))
            try:
                ax.scatter(true_v, true_p, color='tab:blue', marker='o', s=20, linewidths=0.5, edgecolor='black')
            except IndexError:
                pass
            try:
                ax.scatter(false_v, false_p, color='tab:red', marker='o', s=20, linewidths=0.5, edgecolor='black')
            except IndexError:
                pass

            if (y_wt - min(y_true)) < (max(y_true) - y_wt):
                limit_y_true = float(max(y_true) - y_wt)
            else:
                limit_y_true = float(y_wt - min(y_true))
            limit_y_true = limit_y_true * 1.1

            if (y_wt - min(y_pred)) < (max(y_pred) - y_wt):
                limit_y_pred = float(max(y_pred) - y_wt)
            else:
                limit_y_pred = float(y_wt - min(y_pred))
            limit_y_pred = limit_y_pred * 1.1

            plt.vlines(x=(y_wt + limit_y_true) - (((y_wt + limit_y_true) - (y_wt - limit_y_true)) / 2),
                       ymin=y_wt - limit_y_pred, ymax=y_wt + limit_y_pred, color='grey', linewidth=0.5)

            plt.hlines(y=(y_wt + limit_y_pred) - (((y_wt + limit_y_pred) - (y_wt - limit_y_pred)) / 2),
                       xmin=y_wt - limit_y_true, xmax=y_wt + limit_y_true, color='grey', linewidth=0.5)

            crossline = np.linspace(y_wt - limit_y_true, y_wt + limit_y_true)
            plt.plot(crossline, crossline, color='black', linewidth=0.5)

            steps = float(abs(max(y_pred)))
            gradient = []
            for x in np.linspace(0, steps, 100):
                # arr = np.linspace(x/steps, 1-x/steps, steps)
                arr = 1 - np.linspace(x / steps, 1 - x / steps, 100)
                gradient.append(arr)
            gradient = np.array(gradient)

            plt.imshow(gradient, extent=[y_wt - limit_y_true, y_wt + limit_y_true,
                                         y_wt - limit_y_pred, y_wt + limit_y_pred],
                       aspect='auto', alpha=0.8, cmap='coolwarm')  # RdYlGn
            plt.xlim([y_wt - limit_y_true, y_wt + limit_y_true])
            plt.ylim([y_wt - limit_y_pred, y_wt + limit_y_pred])

        plt.xlabel('Measured')
        plt.ylabel('Predicted')
        plt.legend(prop={'size': 8})

        plt.savefig(str(model) + '_' + str(fasta_file[:-6]) + '.png', dpi=500)
    except FileNotFoundError:
        raise FileNotFoundError("Did not find specified model: {}. You can define the threshold of models to be saved;"
                                " e.g. with run_pypef.py run -l LS.fasta -v VS.fasta -t 10.".format(str(model)))

    return ()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Validation modules from Modules_Regression.py
modified for parallelization with Ray.
"""

import matplotlib
matplotlib.use('Agg')
import numpy as np
import os
import ray
import warnings

from pypef.Modules_Regression import full_path, path_aaindex_dir, XY, get_r2

# to handle UserWarning for PLS n_components as error and general regression module warnings
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')


def formatted_output_parallel(aaindex_r2_list, minimum_r2=0.0, no_fft=False):
    """
    takes the sorted list from function R2_List and writes the model names with an R2 ≥ 0
    as well as the corresponding number of components for each model so that the user gets
    a list (Model_Results.txt) of the top ranking models for the given validation set.
    """
    index, value, value2, value3, value4, value5, regr_models, parameters = [], [], [], [], [], [], [], []

    for (idx, val, val2, val3, val4, val5, r_m, pam) in aaindex_r2_list:
        if val >= minimum_r2:
            index.append(idx[:-4])
            value.append('{:f}'.format(val))
            value2.append('{:f}'.format(val2))
            value3.append('{:f}'.format(val3))
            value4.append('{:f}'.format(val4))
            value5.append('{:f}'.format(val5))
            regr_models.append(r_m)
            parameters.append(pam)

    if len(value) == 0:
        raise ValueError('No model with positive R2.')

    data = np.array([index, value, value2, value3, value4, value5, regr_models, parameters]).T
    col_width = max(len(str(value)) for row in data for value in row[:-1]) + 5

    head = ['Index', 'R2', 'RMSE', 'NRMSE', 'Pearson_r', 'Regression', 'Model parameters']
    with open('Model_Results.txt', 'w') as f:
        if no_fft is not False:
            f.write("No FFT used in this model construction, performance"
                    " represents model accuracies on raw encoded sequence data.\n\n")

        heading = "".join(caption.ljust(col_width) for caption in head) + '\n'
        f.write(heading)

        row_length = []
        for row in data:
            row_ = "".join(str(value).ljust(col_width) for value in row) + '\n'
            row_length.append(len(row_))
        row_length_max = max(row_length)
        f.write(row_length_max * '-' + '\n')

        for row in data:
            f.write("".join(str(value).ljust(col_width) for value in row) + '\n')

    return ()


@ray.remote
def parallel(d, core, aa_indices, learning_set, validation_set, regressor='pls', no_fft=False):
    """
    Parallelization of running using the user-defined number of cores.
    Defining the task for each core.
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    aaindex_r2_list = []
    for i in range(d[core][0], d[core][1]):
        aaindex = aa_indices[i]  # Parallelization of AAindex iteration
        xy_learn = XY(full_path(aaindex), learning_set)

        if not no_fft:  # X is FFT-ed of encoded alphabetical sequence
            x_learn, y_learn, _ = xy_learn.get_x_and_y()
        else:  # X is raw encoded of alphabetical sequence
            _, y_learn, x_learn = xy_learn.get_x_and_y()

        # If x_learn (or y_learn) is an empty array, the sequence could not be encoded,
        # because of NoneType value. -> Skip
        if len(x_learn) != 0:
            xy_test = XY(full_path(aaindex), validation_set)

            if not no_fft:  # X is FFT-ed of the encoded alphabetical sequence
                x_test, y_test, _ = xy_test.get_x_and_y()
            else:  # X is the raw encoded of alphabetical sequence
                _, y_test, x_test = xy_test.get_x_and_y()

            r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params = get_r2(
                x_learn, x_test, y_learn, y_test, regressor
            )
            aaindex_r2_list.append([aaindex, r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params])

    return aaindex_r2_list


def r2_list_parallel(learning_set, validation_set, cores, regressor='pls', no_fft=False, sort='1'):
    """
    Parallelization of running using the user-defined number of cores.
    Calling function Parallel to execute the parallel running and
    getting the results from each core each being defined by a result ID.
    """
    aa_indices = [file for file in os.listdir(path_aaindex_dir()) if file.endswith('.txt')]

    split = int(len(aa_indices) / cores)
    last_split = int(len(aa_indices) % cores) + split

    d = {}
    for i in range(cores - 1):
        d[i] = [i * split, i * split + split]

    d[cores - 1] = [(cores - 1) * split, (cores - 1) * split + last_split]

    result_ids = []
    for j in range(cores):  # Parallel running
        result_ids.append(parallel.remote(d, j, aa_indices, learning_set, validation_set, regressor, no_fft))

    results = ray.get(result_ids)

    aaindex_r2_list = []
    for core in range(cores):
        for j, _ in enumerate(results[core]):
            aaindex_r2_list.append(results[core][j])

    try:
        sort = int(sort)
        if sort == 2 or sort == 3:
            aaindex_r2_list.sort(key=lambda x: x[sort])
        else:
            aaindex_r2_list.sort(key=lambda x: x[sort], reverse=True)

    except ValueError:
        raise ValueError("Choose between options 1 to 5 (R2, RMSE, NRMSE, Pearson's r, Spearman's rho.")

    return aaindex_r2_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for creating learning and validation sets
from input CSV file (with value separators sep=',' or sep=';')
having the CSV-format

HEADER_VARIANTS;HEADER_VARIANTS_FITNESS
VARIANT_1;FITNESS_VALUE_1
VARIANT_2;FITNESS_VALUE_2
...

according to the self devised learning and validation set convention
> VARIANT_NAME_1
; FITNESS_1
VARIANT_SEQUENCE_1
> VARIANT_NAME_2
; FITNESS_2
VARIANT_SEQUENCE_2
...
"""

import numpy as np
import random
import pandas as pd
import re
import glob


def get_wt_sequence(sequence_file):
    """
    Gets wild-type sequence from defined input file (can be pure sequence or fasta style)
    """
    # In wild-type sequence .fa file one has to give the sequence of the studied peptide/protein (wild-type)
    # If no .csv is defined is input this tries to find and select any .fa file
    # (risk: could lead to errors if several .fa files exist in folder)
    if sequence_file is None or sequence_file not in glob.glob(sequence_file):
        try:
            types = ('wt_sequence.*', 'wild_type_sequence.*', '*.fa')  # the tuple of file types
            files_grabbed = []
            for files in types:
                files_grabbed.extend(glob.glob(files))
            for i, file in enumerate(files_grabbed):
                if i == 0:
                    sequence_file = file
            if len(files_grabbed) == 0:
                raise FileNotFoundError("Found no input wild-type fasta sequence file (.fa) in current directory!")
            print('Did not find (specified) WT-sequence file! Used wild-type sequence file instead:'
                  ' {}.'.format(str(sequence_file)))
        except NameError:
            raise NameError("Found no input wild-type fasta sequence file (.fa) in current directory!")
    wild_type_sequence = ""
    with open(sequence_file, 'r') as sf:
        for lines in sf.readlines():
            if lines.startswith(">"):
                continue
            lines = ''.join(lines.split())
            wild_type_sequence += lines
    return wild_type_sequence


def csv_input(csv_file):
    """
    Gets input data from defined .csv file (that contains variant names and fitness labels)
    """
    if csv_file is None or csv_file not in glob.glob(csv_file):
        for i, file in enumerate(glob.glob('*.csv')):
            if file.endswith('.csv'):
                if i == 0:
                    csv_file = file
                    print('Did not find (specified) csv file! Used csv input file instead: {}.'.format(str(csv_file)))
        if len(glob.glob('*.csv')) == 0:
            raise FileNotFoundError('Found no input .csv file in current directory.')
    return csv_file


def drop_rows(csv_file, amino_acids, threshold_drop):
    """
    Drops rows from .csv data if below defined fitness threshold or if
    amino acid/variant name is unknown or if fitness label is not a digit.
    """
    separator = ';'
    try:
        df_raw = pd.read_csv(csv_file, sep=separator, usecols=[0, 1])
    except ValueError:
        separator = ','
        df_raw = pd.read_csv(csv_file, sep=separator, usecols=[0, 1])

    label = df_raw.iloc[:, 1]
    sequence = df_raw.iloc[:, 0]

    dropping_rows = []

    for i, row in enumerate(label):
        try:
            row = float(row)
            if row < threshold_drop:
                dropping_rows.append(i)
        except ValueError:
            dropping_rows.append(i)

    for i, variant in enumerate(sequence):
        try:
            if '/' in variant:
                m = re.split(r'/', variant)
                for a, splits in enumerate(m):
                    if splits[0].isdigit() and variant[-1] in amino_acids:
                        continue
                    elif splits[0] not in amino_acids or splits[-1] not in amino_acids:
                        if i not in dropping_rows:
                            dropping_rows.append(i)
                            # print('Does not know this definition of amino acid substitution: Variant:', variant)
            else:
                if variant[0].isdigit() and variant[-1] in amino_acids:
                    continue
                elif variant not in ['wt', 'wild_type']:
                    if variant[0] not in amino_acids or variant[-1] not in amino_acids:
                        dropping_rows.append(i)
                        # print('Does not know this definition of amino acid substitution: Variant:', variant)
        except TypeError:
            raise TypeError('You might consider checking the input .csv for empty first two columns,'
                            ' e.g. in the last row.')

    print('No. of dropped rows: {}.'.format(len(dropping_rows)), 'Total given variants '
                                                                 '(if provided plus WT): {}'.format(len(df_raw)))

    df = df_raw.drop(dropping_rows)
    df.dropna(inplace=True)
    df.reset_index(drop=True, inplace=True)

    return df


def get_variants(df, amino_acids, wild_type_sequence):
    """
    Gets variants and divides and counts the variant data for single substituted and higher substituted variants.
    Raises NameError if variant naming is not matching the given wild-type sequence, e.g. if variant A17C would define
    a substitution at residue Ala-17 to Cys but the wild-type sequence has no Ala at position 17.
    """
    x = df.iloc[:, 0]
    y = df.iloc[:, 1]
    wt_position = None
    single_variants, higher_variants, index_higher, index_lower, higher_values, single_values = [], [], [], [], [], []
    single, double, triple, quadruple, quintuple, sextuple, septuple,\
    octuple, nonuple, decuple, higher = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    for i, variant in enumerate(x):
        if '/' in variant:
            count = variant.count('/')
            if count == 1:
                double += 1
            elif count == 2:
                triple += 1
            elif count == 3:
                quadruple += 1
            elif count == 4:
                quintuple += 1
            elif count == 5:
                sextuple += 1
            elif count == 6:
                septuple += 1
            elif count == 7:
                octuple += 1
            elif count == 8:
                nonuple += 1
            elif count == 9:
                decuple += 1
            else:
                higher += 1
            m = re.split(r'/', variant)
            for a, splits in enumerate(m):
                if splits[0].isdigit() or splits[0] in amino_acids and splits[-1] in amino_acids:
                    new = int(re.findall(r'\d+', splits)[0])
                    if splits[0] in amino_acids:
                        if splits[0] != wild_type_sequence[new - 1]:
                            raise NameError('Position of amino acids in given sequence does not match the given '
                                            'positions in the input data! E.g. see position {} and position {} being {}'
                                            ' in the given sequence'.format(variant, new, wild_type_sequence[new - 1]))
                    higher_var = wild_type_sequence[new - 1] + str(new) + str(splits[-1])
                    m[a] = higher_var
                    if a == len(m) - 1:
                        higher_variants.append(m)
                        if i not in index_higher:
                            index_higher.append(i)
        else:
            single += 1
            if variant.upper() == 'WT' or variant.upper() == 'WILD_TYPE':
                wt_position = i

                continue

            elif variant[0].isdigit() or variant[0] in amino_acids and variant[-1] in amino_acids:
                try:
                    num = int(re.findall(r'\d+', variant)[0])
                except IndexError:
                    raise IndexError('Wrong input format. Please check if the input CSV corresponds to the '
                                     'required input style (while the wild-type protein must be designated as \'WT\').')
                if variant[0] in amino_acids:
                    try:
                        if variant[0] != wild_type_sequence[num - 1]:
                            raise NameError('Position of amino acids in given sequence does not match the given '
                                            'positions in the input data! E.g. see position {} and position {} being {}'
                                            ' in the given sequence.'.format(variant, num, wild_type_sequence[num - 1]))
                    except IndexError:
                        raise IndexError("Found variant sequence position {} in data which "
                                         "is out of range of wild-type sequence length.".format(str(num)))
                try:
                    full_variant = wild_type_sequence[num - 1] + str(num) + str(variant[-1])
                except IndexError:
                    raise IndexError("Found variant sequence position {} in data which "
                                     "is out of range of wild-type sequence length.".format(str(num)))
                single_variants.append([full_variant])
                if i not in index_lower:
                    index_lower.append(i)
    print('Single (for mklsvs if provided plus WT): {}.'.format(single), 'Double: {}.'.format(double),
          'Triple: {}.'.format(triple), 'Quadruple: {}.'.format(quadruple), 'Quintuple: {}.'.format(quintuple),
          'Sextuple: {}.'.format(sextuple), 'Septuple: {}.'.format(septuple), 'Octuple: {}.'.format(octuple),
          'Nonuple: {}.'.format(nonuple), 'Decuple: {}.'.format(decuple), 'Higher: {}.'.format(higher))
    for vals in y[index_higher]:
        higher_values.append(vals)
    for vals in y[index_lower]:
        single_values.append(vals)
    if wt_position is not None:
        single_variants.append(['WT'])
        single_values.append(y[wt_position])

    single_variants, single_values = tuple(single_variants), tuple(single_values)
    higher_variants, higher_values = tuple(higher_variants), tuple(higher_values)

    return single_variants, single_values, higher_variants, higher_values


def make_sub_ls_vs(single_variants, single_values, higher_variants, higher_values, directed_evolution=False):
    """
    Creates learning and validation sets, fills learning set with single substituted variants and splits
    rest (higher substituted) for learning and validation sets: 3/4 to LS and 1/4 to VS
    """
    print('No. of single substituted variants (if provided plus WT): {}.'.format(len(single_variants)),
          'No. of values: {}'.format(len(single_values)))
    print('No. of higher substituted variants: {}.'.format(len(higher_variants)),
          'No. of values: {}'.format(len(higher_values)))

    if len(single_values) != len(single_variants):
        print('Error due to different lengths for given variants and label!'
              ' No. of single substituted variants: {}.'.format(len(single_variants)),
              ' Number of given values: {}.'.format(len(single_values)))

    if len(higher_values) != len(higher_variants):
        print('Error due to different lengths for given variants and label! No. of higher subst. variants: {}.'
              .format(len(higher_variants)), ' Number of given values: {}.'.format(len(higher_values)))

    # 1. CREATION OF LS AND VS SPLIT FOR SINGLE FOR LS AND HIGHER VARIANTS FOR VS
    sub_ls = list(single_variants)  # Substitutions of LS
    val_ls = list(single_values)    # Values of LS

    sub_vs = []                     # Substitutions of VS
    val_vs = []                     # Values of VS

    if directed_evolution is False:
        for i in range(len(higher_variants)):
            if len(higher_variants) < 6:  # if less than 6 higher variants all higher variants are appended to VS
                sub_vs.append(higher_variants[i])
                val_vs.append(higher_values[i])
            elif (i % 3) == 0 and i != 0:  # 1/4 of higher variants to VS, 3/4 to LS: change here for LS/VS ratio change
                sub_vs.append(higher_variants[i])
                val_vs.append(higher_values[i])
            else:                       # 3/4 to LS
                sub_ls.append(higher_variants[i])
                val_ls.append(higher_values[i])

    return sub_ls, val_ls, sub_vs, val_vs


def make_sub_ls_vs_randomly(single_variants, single_values, higher_variants, higher_values):
    """
    Creation of learning set and validation set by randomly splitting sets
    """
    length = len(single_variants) + len(higher_variants)
    range_list = np.arange(0, length)

    vs = []
    ls = []
    while len(ls) < length * 4 // 5:
        random_num = random.choice(range_list)
        if random_num not in ls:
            ls.append(random_num)

    for j in range_list:
        if j not in ls:
            vs.append(j)

    combined = single_variants + higher_variants  # substitutions
    combined2 = single_values + higher_values  # values

    sub_ls = []
    val_ls = []
    tot_sub_ls, tot_val_ls = [], []
    tot_sub_vs, tot_val_vs = [], []

    for i in ls:
        sub_ls.append(combined[i])
        val_ls.append(combined2[i])

    sub_vs = []
    val_vs = []
    for j in vs:
        sub_vs.append(combined[j])
        val_vs.append(combined2[j])

    for subs in sub_ls:
        for subs2 in sub_vs:
            if subs == subs2:
                print('\n<Warning> LS and VS overlap for: {} - You might want to consider checking the provided '
                      'datasets for multiple entries'.format(subs), end=' ')

    tot_sub_ls.append(sub_ls)
    tot_val_ls.append(val_ls)
    tot_sub_vs.append(sub_vs)
    tot_val_vs.append(val_vs)

    return tot_sub_ls[0], tot_val_ls[0], tot_sub_vs[0], tot_val_vs[0]


def make_fasta_ls_vs(filename, wt, sub, val):  # sub = substitution, val = value
    """
    Creates learning and validation sets (.fasta style files)
    """
    myfile = open(filename, 'w')
    for i, var in enumerate(sub):  # var are lists of (single or multiple) substitutions
        temp = list(wt)
        name = ''
        separation = 0
        if var == ['WT']:
            name = 'WT'
        else:
            for single_var in var:  # single entries of substitution list
                position_index = int(str(single_var)[1:-1]) - 1
                new_amino_acid = str(single_var)[-1]
                temp[position_index] = new_amino_acid
                # checking if multiple entries are inside list
                if separation == 0:
                    name += single_var
                else:
                    name += '/' + single_var
                separation += 1
        print('>', name, file=myfile)
        print(';', val[i], file=myfile)
        print(''.join(temp), file=myfile)
    myfile.close()



from setuptools import setup, find_packages
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r", encoding="utf-8") as fh_single_core:
    with open("requirements_parallelization.txt", "r", encoding="utf-8") as fh_multithreading:
        requirements = fh_single_core.read() + fh_multithreading.read()

setup(
    name='pypef',
    version='0.1.3',
    author='Niklas Siedhoff & Alexander-Maurice Illig',
    author_email='n.siedhoff@biotec.rwth-aachen.de',
    license='CC BY-NC-SA 4.0',
    description='A CLI tool (not intended for API use) for performing data-driven protein engineering by building '
                'machine learning models from sequence variant-fitness data (e.g., provided as CSV data).',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/niklases/PyPEF',
    py_modules=['pypef'],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[requirements],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    entry_points={
        'console_scripts': [
            'pypef=pypef.run_pypef:run'
        ],
    }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @author: Niklas Siedhoff, Alexander-Maurice Illig
# <n.siedhoff@biotec.rwth-aachen.de>, <a.illig@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for performing random evolution walks
similar as presented by Biswas et al.
"""

import os
import random
import matplotlib.pyplot as plt
import numpy as np
import warnings

from pypef.Modules_Regression import predict

# ignoring warnings of scikit-learn regression
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')


def mutate_sequence(seq, m, model, prev_mut_loc, aas, sub_ls, iteration, counter, usecsv, csvaa):
    """
    produces a mutant sequence (integer representation), given an initial sequence
    and the number of mutations to introduce ("m") for in silico directed evolution
    """
    try:
        os.mkdir('EvoTraj')
    except FileExistsError:
        pass
    myfile = 'EvoTraj/' + str(model) + '_EvoTraj_' + str(counter+1) + '_DEiter_' + str(iteration+1) + '.fasta'
    var_seq_list = []
    with open(myfile, 'w') as mf:
        for i in range(m):  # iterate through number of mutations to add
            rand_loc = random.randint(prev_mut_loc - 8, prev_mut_loc + 8)  # find random position to mutate
            while (rand_loc <= 0) or (rand_loc >= len(seq)):
                rand_loc = random.randint(prev_mut_loc - 8, prev_mut_loc + 8)

            if usecsv is True:  # Only perform directed evolution on positional csv variant data
                pos_list = []
                aa_list = []
                for aa_positions in sub_ls:
                    for pos in aa_positions:
                        pos_int = int(pos[1:-1])
                        if pos_int not in pos_list:
                            pos_list.append(pos_int)
                        if csvaa is True:
                            new_aa = str(pos[-1:])
                            if new_aa not in aa_list:
                                aa_list.append(new_aa)
                            aas = aa_list
                # Select closest Position to single AA positions
                absolute_difference_function = lambda list_value: abs(list_value - rand_loc)
                try:
                    closest_loc = min(pos_list, key=absolute_difference_function)
                except ValueError:
                    raise ValueError("No positions for recombination found. Likely no single "
                                     "substitutional variants found in provided .csv file.")
                rand_loc = closest_loc - 1   # - 1 as Position 17 is 16 when starting with 0 index
            rand_aa = random.choice(aas)  # find random amino acid to mutate to
            sequence = seq
            seq_ = list(sequence)
            seq_[rand_loc] = rand_aa  # update sequence to have new amino acid at randomly chosen position
            seq_ = ''.join(seq_)
            var = str(rand_loc+1)+str(rand_aa)
            var_seq_list.append([var, seq_])

            print('> {}{}'.format(rand_loc+1, rand_aa), file=mf)
            print(''.join(seq_), file=mf)

    return dict(var_seq_list)   # chose dict as one can easily change to more sequences to predict per iteration


def restructure_dict(prediction_dict):
    """
    Exchange "key and value" of variant-fitness "tuple dictionary"
    """
    restructured = []
    for pred in prediction_dict:
        restruct = []
        for p in pred:
            restruct.insert(0, p)
        restructured.append(restruct)
    structured_dict = dict(restructured)
    return structured_dict


def write_mcmc_predictions(model, iteration, predictions, counter):
    """
    write predictions to EvoTraj folder to .fasta files for each iteration of evolution
    """
    with open('EvoTraj/' + str(model) + '_EvoTraj_' + str(counter+1) + '_DEiter_' + str(iteration+1)
              + '.fasta', 'r') as f_in:
        with open('EvoTraj/' + str(model) + '_EvoTraj_' + str(counter+1) + '_DEiter_' + str(iteration+1)
                  + '_prediction.fasta', 'w') as f_out:
            for line in f_in:
                f_out.write(line)
                if '>' in line:
                    key = line[2:].strip()
                    f_out.writelines('; ' + str(predictions.get(key)) + '\n')
    return ()


def in_silico_de(s_wt, num_iterations, model, amino_acids, temp, path, sub_ls, counter,
                 no_fft=False, negative=False, usecsv=False, csvaa=False, print_matrix=False):
    """
    Perform directed evolution by randomly selecting a sequence position for substitution and randomly choose the
    amino acid to substitute to. New sequence gets accepted if meeting the Metropolis criterion and will be
    taken for new substitution iteration.
    Metropolis-Hastings-driven directed evolution, similar to Biswas et al.:
    Low-N protein engineering with data-efficient deep learning,
    see https://github.com/ivanjayapurna/low-n-protein-engineering/tree/master/directed-evo
    """
    v_traj = []  # initialize an array to keep records of the variant names for this trajectory
    y_traj = []  # initialize an array to keep records of the fitness scores for this trajectory
    s_traj = []  # initialize an array to keep records of the protein sequences for this trajectory

    # iterate through the trial mutation steps for the directed evolution trajectory
    for i in range(num_iterations):  # num_iterations
        if i == 0:  # get first v, y, s
            # randomly choose the location of the first mutation in the trajectory
            mut_loc_seed = random.randint(0, len(s_wt))
            # m = 1 instead of (np.random.poisson(2) + 1)
            var_seq_dict = mutate_sequence(s_wt, 1, model, mut_loc_seed, amino_acids, sub_ls, 0, counter, usecsv, csvaa)

            predictions = predict(path, 'EvoTraj/' + str(model) + '_EvoTraj_' + str(counter+1) + '_DEiter_'
                                  + str(i+1) + '.fasta', model, None, no_fft, print_matrix)
            predictions = restructure_dict(predictions)

            write_mcmc_predictions(model, i, predictions, counter)

            ys, variants = [], []
            for var in predictions:
                variants.append(var)
                ys.append(predictions.get(var))

            y, var = ys[0], variants[0]  # only one entry anyway
            new_mut_loc = int(var[:-1]) - 1
            sequence = var_seq_dict.get(var)

            v_traj.append(var)
            y_traj.append(y)
            s_traj.append(sequence)

        else:  # based on first v, y, s go deeper in mutations --> new_v, new_y, new_s if accepted
            # only chose 1 mutation to introduce and not:
            # mu = np.random.uniform(1, 2.5) --> Number of Mutations = m = np.random.poisson(mu - 1) + 1
            new_var_seq_dict = mutate_sequence(sequence, 1, model, new_mut_loc, amino_acids,
                                               sub_ls, i, counter, usecsv, csvaa)
            predictions = predict(path, 'EvoTraj/' + str(model) + '_EvoTraj_' + str(counter+1) + '_DEiter_'
                                  + str(i+1) + '.fasta', model, None, no_fft, print_matrix)
            predictions = restructure_dict(predictions)

            write_mcmc_predictions(model, i, predictions, counter)

            new_ys, new_variants = [], []
            for var in predictions:
                new_variants.append(var)
                new_ys.append(predictions.get(var))

            new_y, new_y_var = new_ys[0], new_variants[0]
            new_mut_loc = int(new_y_var[:-1]) - 1
            new_sequence = new_var_seq_dict.get(new_y_var)

            # probability function for trial sequence
            # The lower the fitness (y) of the new variant, the higher are the chances to get excluded
            with warnings.catch_warnings():  # catching Overflow warning
                warnings.simplefilter("ignore")
                try:
                    boltz = np.exp(((new_y - y) / temp), dtype=np.longfloat)
                    if negative is True:
                        boltz = np.exp((-(new_y - y) / temp), dtype=np.longfloat)
                except OverflowError:
                    boltz = 1
            p = min(1, boltz)
            rand_var = random.random()  # random float between 0 and 1
            if rand_var < p:  # Metropolis-Hastings update selection criterion
                # print('Updated sequence as: Rand ({}) < Boltz ({})'.format(str(rand_var), str(boltz)))
                # print(str(new_mut_loc + 1) + " " + sequence[new_mut_loc] + "->" + new_sequence[new_mut_loc])
                var, y, sequence = new_y_var, new_y, new_sequence  # if criteria is met, update sequence and
                                                                   # corresponding fitness
                v_traj.append(var)  # update the variant naming trajectory records for this iteration of mutagenesis
                y_traj.append(y)  # update the fitness trajectory records for this iteration of mutagenesis
                s_traj.append(sequence)  # update the sequence trajectory records for this iteration of mutagenesis

    return v_traj, s_traj, y_traj


def run_de_trajectories(s_wt, model, y_wt, num_iterations, num_trajectories, de_record_folder, amino_acids, temp, path,
                        sub_ls, no_fft=False, negative=False, save=False, usecsv=False, csvaa=False,
                        print_matrix=False):
    """
    Runs the directed evolution by adressing the in_silico_de function and plots the evolution trajectories.
    """
    v_records = []  # initialize list of sequence variant names
    s_records = []  # initialize list of sequence records
    y_records = []  # initialize list of fitness score records

    for i in range(num_trajectories):  # iterate through however many mutation trajectories we want to sample
        # call the directed evolution function, outputting the trajectory sequence and fitness score records
        v_traj, s_traj, y_traj = in_silico_de(s_wt, num_iterations, model, amino_acids, temp, path, sub_ls, i,
                                              no_fft, negative, usecsv, csvaa, print_matrix)
        v_records.append(v_traj)  # update the variant naming trajectory records for this full mutagenesis trajectory
        s_records.append(s_traj)  # update the sequence trajectory records for this full mutagenesis trajectory
        y_records.append(y_traj)  # update the fitness trajectory records for this full mutagenesis trajectory

        if save:
            try:
                os.mkdir(de_record_folder)
            except FileExistsError:
                pass
            # save sequence records for trajectory i
            np.savetxt(de_record_folder + "/" + str(model) + "_trajectory" + str(i+1)
                       + "_seqs.txt", np.array(s_traj), fmt="%s")
            # save fitness records for trajectory i
            np.savetxt(de_record_folder + "/" + str(model) + "_trajectory" + str(i+1)
                       + "_fitness.txt", np.array(y_traj))

    # numpy warning filter needed for arraying ragged nested sequences
    np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
    v_records = np.array(v_records)
    s_records = np.array(s_records)
    y_records = np.array(y_records)

    fig, ax = plt.subplots()  # figsize=(10, 6)
    ax.locator_params(integer=True)
    f_len_max = 0
    for j, fitness_array in enumerate(y_records):
        if y_wt is not None:
            f_len = len(fitness_array)
            if f_len > f_len_max:
                f_len_max = f_len
            ax.plot(np.arange(1, len(fitness_array)+2, 1), np.insert(fitness_array, 0, y_wt),  # insert y_WT at Pos. 0
                    '-o', alpha=0.7, markeredgecolor='black', label='EvoTraj' + str(j+1))
        else:
            ax.plot(np.arange(1, len(fitness_array)+1, 1), fitness_array,
                    '-o', alpha=0.7, markeredgecolor='black', label='EvoTraj' + str(j+1))

    label_x_y_name = []
    for k, l in enumerate(v_records):  # k = 1, 2, 3, .., ; l = variant label array
        for kk, ll in enumerate(l):  # kk = 1, 2, 3, ...  (=x); ll = variant label; y_records[k][kk] = fitness (=y)
            if y_wt is not None:     # kk+2 as enumerate starts with 0 and WT is 1 --> start labeling with 2
                label_x_y_name.append(ax.text(kk+2, y_records[k][kk], ll))
            else:
                label_x_y_name.append(ax.text(kk+1, y_records[k][kk], ll))

    # adjusting variant text labels
    from adjustText import adjust_text
    adjust_text(label_x_y_name, only_move={'points': 'y', 'text': 'y'}, force_points=0.5)
    leg = ax.legend()
    if y_wt is not None:
        # plt xticks with locations and labels following: xticks(locs, labels)
        if f_len_max > 10:
            plt.xticks(np.arange(1, f_len_max + 2, 5), ['WT'] + ((np.arange(5, f_len_max + 1, 5)).tolist()))
        else:
            plt.xticks(np.arange(1, f_len_max + 2, 1), ['WT'] + ((np.arange(1, f_len_max + 1, 1)).tolist()))

    plt.ylabel('Predicted fitness')
    plt.xlabel('Mutation trial steps')
    plt.savefig(str(model) + '_DE_trajectories.png', dpi=500)

    return s_records, y_records
